import {App, Component, inject, Plugin} from "vue"
import {DynamicModalOptions, VueFinalModalProperty} from "vue-final-modal"
import {RouteLocationNormalized, Router} from "vue-router"

type ClosableModal = {
  props: { name: string }
  toggle: () => void
  params: { value: { name: string, _modalName: string } }
}

type ScriptSetupComponent = Component & {
  __name: string // <script setup> 을 사용하는 경우 __name 속성으로 컴포넌트 이름이 들어간다.
}

// DynamicModalOptions 객체로부터 modal name 을 추출한다.
const getModalName = (modal: DynamicModalOptions): string => {
  const name = (typeof modal.component === 'string')
      ? modal.component
      : modal.component?.name || (modal.component as ScriptSetupComponent)?.__name || ''
  if (!name) throw Error(`Component name is missing. component = ${modal.component}`)
  return name
}

type FullPathOnShow = {
  before: string
  after: string
}

export type HistoryModalPluginOption = {
  modal: VueFinalModalProperty
  router: Router
}

type OpenHistoryModalInfo = {
  name: string
  fullPathOnShow: FullPathOnShow // show() 시점의 route.fullPath
}

const historyModalKey = Symbol('HistoryModalPluginKey')

/**
 * 반드시 `setup()` 함수 내에서 호출되어야 한다.
 */
export const useHistoryModal = (): HistoryModal => {
  const manager = inject<HistoryModal>(historyModalKey)
  if (!manager) {
    throw new Error('HistoryModalPlugin is not installed!')
  }
  return manager
}

export interface HistoryModal {

  show(option: DynamicModalOptions, params?: Record<string, unknown>): Promise<void>

  hide(modalName: string): Promise<void>

  hideAllModals(): Promise<void>

  getOpenModalKeys(): string[]
}

class HistoryModalImpl implements HistoryModal {

  private unregisterRouterBeforeEach?: () => void
  private openModals: Map<string, OpenHistoryModalInfo> = new Map()

  constructor(public readonly modal: VueFinalModalProperty,
              public readonly router: Router) {
    this.router.isReady().then(() => this.initialize())
  }

  private async initialize(): Promise<void> {
    // router 이동할 때마다 modal fragment 를 제거한다.
    this.unregisterRouterBeforeEach = this.router.beforeEach(
        (to: RouteLocationNormalized, from: RouteLocationNormalized): true => {

          if (from.hash !== to.hash) { // hash 로 historyModal 을 표시하기 때문에, hash 변경은 별도로 처리한다.
            // modal fragment 가 있다가 없어졌으면, 해당 modal 을 닫는다.
            if ((from.hash.length > 1) && (to.hash.length < from.hash.length)) {
              const fromFrags = this.getModalFragments(from.hash)
              const toFrags = this.getModalFragments(to.hash)
              const removed = fromFrags.filter(f => !toFrags.includes(f))
              removed.forEach(name => this.hide(name))
            }
          } else if (from.fullPath !== to.fullPath) {
            // hash 이외의 fullPath 가 변경되었다면, modal 을 모두 닫는다.
            this.modal.hideAll()
          }
          return true
        })
    await this.removeAllModalFragments()
  }

  private async removeAllModalFragments(): Promise<void> {
    const route = this.router.currentRoute.value
    try {
      await this.router.replace({query: route.query, hash: ''})
    } catch (err) {
      console.log(err)
    }
  }

  private getModalFragments(hash: string): string[] {
    const curHash = hash.startsWith('#') ? hash.substring(1) : ''
    return (curHash.length > 0) ? curHash.split(',') : []
  }

  private async removeModalFragment(modalName: string): Promise<void> {
    const route = this.router.currentRoute.value
    const targetModal = this.openModals.get(modalName)

    if (targetModal?.fullPathOnShow?.after === route.fullPath) {
      // fullPathOnShow.after 가 현재 route.fullPath 와 같다면, x 버튼 눌러 modal 을 닫은 것이다.
      this.router.back()

    } else if (targetModal?.fullPathOnShow?.before === route.fullPath) {
      // fullPathOnShow.before 가 현재 route.fullPath 와 같다면, 이미 back 을 누른 것이다.

    } else {
      // 그렇지 않으면 hash 를 제거한다.
      const curFrags = this.getModalFragments(route.hash)
      const newFrags = curFrags.filter(f => f !== modalName).join(',')
      const newHash = newFrags.length > 0 ? `#${newFrags}` : ''
      if (route.hash !== newHash) {
        await this.router.replace({query: route.query, hash: newHash})
      }
    }
  }

  private async appendModalFragment(modalName: string): Promise<void> {
    const route = this.router.currentRoute.value
    const curFrags = this.getModalFragments(route.hash)
    if (!curFrags.includes(modalName)) { // hash 가 없는 경우에만 추가.
      curFrags.push(modalName)
    }
    const newHash = '#' + curFrags.join(',')
    await this.router.push({query: route.query, hash: newHash})
  }

  public async show(option: DynamicModalOptions, params?: Record<string, unknown>): Promise<void> {
    const modalName = getModalName(option)

    // close handler 를 추가한다.
    option.on = {...option.on, closed: async () => await this.handleClose(modalName)}

    const newParams = {_modalName: modalName, ...params}
    await this.modal.show(option, newParams)

    const route = this.router.currentRoute.value

    await this.appendModalFragment(modalName)

    this.openModals.set(modalName, {
      name: modalName,
      fullPathOnShow: {
        before: route.fullPath,
        after: this.router.currentRoute.value.fullPath
      }
    })
  }

  public async hide(modalName: string): Promise<void> {
    this.modal.openedModals
        .filter(open => {
          const name = (open as unknown as ClosableModal).params?.value?._modalName
          return name === modalName
        })
        .forEach(toHide => (toHide as unknown as ClosableModal).toggle())
    this.openModals.delete(modalName)
  }

  private async handleClose(modalName: string): Promise<void> {
    await this.removeModalFragment(modalName)
  }

  public async hideAllModals(): Promise<void> {
    for (const modalName of this.openModals.keys()) {
      await this.hide(modalName)
    }
  }

  public getOpenModalKeys(): string[] {
    return Array.from(this.openModals.keys())
  }
}

/**
 * ```ts
 * createApp(GuideMain)
 *     .use(router)
 *     .use(ModalPlugin)
 *     .use(HistoryModalPlugin, {router, modal: ModalPlugin.modal})
 * ```
 */
export const HistoryModalPlugin: Plugin = {

  install(app: App, option?: HistoryModalPluginOption): void {
    if (!option) {
      throw new Error("option (HistoryModalPluginOptions) is missing!")
    }
    if (!option.modal) {
      throw new Error('option.modal is missing!')
    }
    if (!option.router) {
      throw new Error('option.router is missing!')
    }

    const manager = new HistoryModalImpl(option.modal, option.router)
    app.provide(historyModalKey, manager)
  }
}

/**
 * historyModal 이 열려있으면 router.back() 후 action 을 실행하고,
 * 그렇지 않으면 곧바로 action 을 실행한다.
 * @param historyModal
 * @param router
 * @param action
 */
export const closeHistoryModalAndDo = async (historyModal: HistoryModal,
                                             router: Router,
                                             action: () => unknown): Promise<void> => {
  const open = historyModal.getOpenModalKeys()
  if (open.length > 0) {
    await router.back()
    setTimeout(action, 100)
  } else {
    action()
  }
}